# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Any, Optional

import torch
import torch.nn.functional as F

from ..models.utils import ModelOutput


class SegmentationPostprocessor:
    def __init__(self, conf_model):
        pass
    def __call__(self, outputs: ModelOutput, original_shape):
        pred = outputs['pred']
        H, W = original_shape[-2:]
        # upsample logits to the images' original size
        pred = F.interpolate(
            pred, size=(H, W), mode="bilinear", align_corners=False
        )
        pred = torch.max(pred, dim=1)[1]  # argmax
        return pred.detach().cpu().numpy()
